# -*- coding: utf-8 -*-
import os
import logging

from django.http.response import HttpResponse
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt
from django.utils import timezone
from django.template import loader

from odk.models import XFormSubmit
from odk.utils import ManageFile

LOG_DEBUG = logging.getLogger("mydebug")


class BaseOpenRosaResponse(HttpResponse):
    """
    OpenRosa settings
    """
    status_code = 201
    def __init__(self, *args, **kwargs):
        super(BaseOpenRosaResponse, self).__init__(*args, **kwargs)
        self.setdefault('X-OpenRosa-Version', '1.0')
        self.setdefault('Date', timezone.now().strftime('%a, %d %b %Y %H:%M:%S %Z'))
        self.setdefault('X-OpenRosa-Accept-Content-Length', 10000000)
        self.setdefault('Content-Type', 'text/xml; charset=utf-8')


@csrf_exempt
def xform_submit(request, username=None):
    """
    Function used to manage the submissions and the responses. 
    """    
    if request.method == 'HEAD':
        """
        We are waiting the HEAD request, that is sent by ODK Collect
        to check the server configuration before sending the POST request.
        Collect expects 1 information as reponse to a HEAD request:
        Status code = 204
        """
        response = BaseOpenRosaResponse(status=204)
        # Keep base_url and add /submission
        response['Location'] = request.build_absolute_uri().replace(
            request.get_full_path(), '/odk/submission')
        return response
        
    
    if request.method == 'POST':
        """
        Just save submitted xml forms in overwrite_storage
        ODK needs two things for a form to be considered successful
        1. status code 201 (created).
        2. location header with absolute_uri
        """
        data = {}
    
        subfile = ManageFile(request)
        msg, xsub = subfile.process_xformsubmit()

        data['message'] = msg
        data['encrypted'] = False

        if xsub is None:
            status_code = 404
        else:
            status_code = 201
            data['formid'] = xsub.form_id
            data['instanceID'] = xsub.instanceid
            data['submissionDate'] = xsub.submitted_on
            data['markedAsCompleteDate'] = xsub.submitted_on

        # Building response with XML template and data dict
        response = BaseOpenRosaResponse(loader.get_template('submission.xml').render(data))
        response.status_code = status_code

    return response





# Generated by Django 3.2.13 on 2022-06-27 12:13

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import odk.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='XFormSubmit',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('form_id', models.SlugField(editable=False, help_text='Retrieved from XML form (settings tab of Excel file)', max_length=200)),
                ('version', models.CharField(editable=False, help_text='Retrieved from XML form (settings tab of Excel file)', max_length=200)),
                ('instanceid', models.UUIDField(unique=True)),
                ('deviceid', models.CharField(blank=True, max_length=255, null=True)),
                ('survey_date', models.DateField(blank=True, null=True, verbose_name='Encoding date')),
                ('picture_files', models.JSONField(blank=True, null=True, verbose_name='Picture file names')),
                ('xml_file', models.FileField(blank=True, help_text='XML file sended through ODK Collect Mobile App', upload_to='', verbose_name='Submitted form')),
                ('xml_content', models.TextField(help_text='XML content sended through ODK Collect Mobile App', verbose_name='Content of XML file submitted')),
                ('submitted_by', models.CharField(blank=True, max_length=255, null=True, verbose_name='Submitted by')),
                ('submitted_on', models.DateTimeField(auto_now_add=True, verbose_name='Submitted on')),
                ('inserted_on', models.DateTimeField(blank=True, null=True, verbose_name='Inserted on')),
            ],
            options={
                'verbose_name': 'Submitted form',
                'verbose_name_plural': 'Submitted forms',
                'ordering': ('submitted_on',),
            },
        ),
        migrations.CreateModel(
            name='XForm',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('xls_file', models.FileField(blank=True, help_text='XLSForm with 3 tabs: survey, choices, settings', null=True, upload_to=odk.models.xform_path, verbose_name='Excel file')),
                ('xml_file', models.FileField(blank=True, help_text="XLSForm converted by <a href='https://getodk.org/xlsform/' target='_blank'>https://getodk.org/xlsform/</a>", null=True, upload_to=odk.models.xform_path, verbose_name='XML file')),
                ('xml_content', models.TextField(blank=True, null=True, verbose_name='Content of XML form')),
                ('form_id', models.SlugField(editable=False, help_text='Retrieved from XLSForm (xls settings tab)', max_length=200)),
                ('version', models.CharField(editable=False, help_text='Retrieved from XLSForm (xls settings tab)', max_length=200)),
                ('title', models.CharField(editable=False, help_text='Retrieved from XLSForm (xls settings tab)', max_length=250, verbose_name='Title')),
                ('short_desc', models.CharField(blank=True, max_length=250, verbose_name='Short description')),
                ('created_on', models.DateTimeField(auto_now_add=True, verbose_name='Created on')),
                ('model_created_on', models.DateTimeField(blank=True, null=True, verbose_name='Model created on')),
                ('created_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='xform_created_by', to=settings.AUTH_USER_MODEL, verbose_name='Created by')),
            ],
            options={
                'verbose_name': 'Available form',
                'ordering': ('xml_file',),
                'unique_together': {('form_id', 'version')},
            },
        ),
    ]



def rm_digit(mystr):
    """remove all digit from a string"""
    strnew = ''.join([i for i in mystr if not i.isdigit()])
    return strnew


def convert2camelcase(word):
    """
    convert a string with underscore (package naming convention) to CamelCase
    ex: hello_world => HelloWorld
    """
    return ''.join(x.capitalize() or '_' for x in word.split('_'))


import os
import glob
import logging

from django.core.management import call_command
from xlsconv.django import xls2django
from config.settings import MEDIA_ROOT
from odk.models import XForm
from odkdata.utils import rm_digit, convert2camelcase


odkdata_dir = os.path.dirname(__file__)


LOG = logging.getLogger(__name__)



def get_last_file():
    """
    return
    xls_path: "/var/www/projectname/media/XForm/pdt9.xlsx"
    xls_name: "pdt9.xlsx"
    xls_root: "pdt"
    Not used for the moment (should check form_id inside xml)
    """
    xform_path = os.path.join(MEDIA_ROOT, 'XForm/')
    xlsx_path = f"{xform_path}*.xlsx"
    file_list = glob.glob(xlsx_path)

    xls_path = max(file_list, key=os.path.getctime)
    xls_name = xls_path.replace(xform_path, '').replace('.xlsx', '')
    xls_root = rm_digit(xls_name).replace('.xlsx', '')

    return xls_path, xls_name, xls_root


def get_xlsx_path(xform):
    """Not used"""
    if xform is None:
        xls_path, xls_name, xls_root = get_last_file()
    else:
        xls_path = xform.xls_file.path
        xls_name = xform.xls_file.name
        xls_root = rm_digit(xls_name).replace('.xlsx', '')

    LOG.warning(f"xls_path: {xls_path}, xls_root: {xls_root}")
    return xls_path, xls_name, xls_root
    

def append_init_file(pkg_name):
    """
    Check init content and add new model declaration if absent
    """
    model_name = convert2camelcase(pkg_name)
    path = os.path.join(odkdata_dir, 'models', '__init__.py')
    newline = f"from odkdata.models.{pkg_name} import {model_name}\n"

    try:
        with open(path, "r") as init_file_r:
            lines = init_file_r.readlines()
    except (IOError, OSError) as e:
        LOG.error(e)

    if newline not in lines:
        with open(path, "a+") as init_file:
            init_file.write(newline)


def model_correction(pkg_path, form_id):
    """
    Loop through model created by xlsconv.django.xls2django
    an rewrite it with 3 changes
    1. rename all old_names > new_names (ex: Pdt9 become Pdt)
    2. create UUID field instanceid
    3. add "null=True, blank=True" for srid field
    """
    old_name = convert2camelcase(form_id)
    new_name = convert2camelcase(rm_digit(form_id))

    path_old = pkg_path
    path_new = f"{odkdata_dir}/models/{rm_digit(form_id)}.py"
    sp4 = "    "
    try:
        with open(path_old, "r") as f_old, open(path_new, "w") as f_new:
            for line in f_old:
                # replace all old_name to new_name
                line = line.replace(old_name, new_name).replace(old_name.lower(), new_name.lower())
                f_new.write(line)

                if "from django.contrib.gis.db import models" in line:
                    f_new.write("from odk.models import XFormSubmit")
                if f"class {new_name}(models.Model):\n" in line:
                    f_new.write(f"{sp4}xfs = models.OneToOneField(XFormSubmit, on_delete=models.CASCADE, primary_key=True)\n")
                    f_new.write(f"{sp4}instanceid = models.UUIDField(unique=True)\n")
                if f'verbose_name_plural = "{new_name.lower()}s"' in line:
                    f_new.write(f'{sp4}{sp4}ordering = ["-xfs__submitted_on"]\n')
                    f_new.write("\n")

                    f_new.write('    def get_formid_version(self):\n')
                    f_new.write('        return f"{self.xfs.form_id}v{self.xfs.version}"')
                    f_new.write('\n\n')

                    f_new.write('    def get_submit_date(self):\n')
                    f_new.write('''        return f"{self.xfs.submitted_on.strftime('%Y-%m-%d %H:%M:%S')}"''')
                    f_new.write('\n\n')

                    f_new.write('    def get_title(self):\n')
                    f_new.write('''        return f"{self.xfs.form_id} date: {self.today.strftime('%Y-%m-%d')}"''')
                    f_new.write('\n\n')

                    f_new.write('    def __str__(self):\n')
                    f_new.write('''        return f"{self.xfs.form_id} @ {self.xfs.submitted_on.strftime('%Y-%m-%d %H:%M:%S')}"''')
                    f_new.write('\n\n')

    except (IOError, OSError) as e:
        LOG.error(e)
        return False

    return True


def process(xform=None):
    """
    param: XForm instance
    return Boolean
    """
    if xform is None:
        LOG.warning("You must provide a XForm instance")
        return False
        
    # 1. Init vars
    xls_path = xform.xls_file.path
    pkg_name = rm_digit(xform.form_id)
    
    pkg_path = os.path.join(odkdata_dir, 'models', f"{pkg_name}_orig.py")

    # 2. xls2django
    fpointer = xls2django(xls_path)
    try:
        with open(pkg_path, "w") as fmodel:
            print(fpointer, file=fmodel)
    except (IOError, OSError) as e:
        LOG.error(e)
        return False

    # 3. Append_init
    try:
        append_init_file(pkg_name)
    except Exception as xcpt:
        LOG.error(xcpt)
        return False
    
    # 4. Model correction
    if not model_correction(pkg_path, xform.form_id):
        return False

    return True


def apply_changes2db():
    """
    function calling
    makemigrations and migrate
    Works fine but no feedback to user => Leave from now 2022-06-16
    """
    try:
        call_command('makemigrations', 'odkdata')
    except Exception as xcpt:
        LOG.error(xcpt)
        return False

    try:
        call_command('migrate')
    except Exception as xcpt:
        LOG.error(xcpt)
        return False
    return True


# -*- coding: utf-8 -*-
import os
import logging

from braces.views import LoginRequiredMixin
# django modules
from django.utils.translation import gettext as _
from django.utils.timezone import now
from django.contrib.auth.decorators import login_required
from django.urls import reverse_lazy
from django.views import generic
from django.shortcuts import render, redirect
from django.contrib import messages
from django.http import HttpResponseRedirect
# odk modules
from .storage import overwrite_storage
from .models import XForm, XFormSubmit, xform_path
from .forms import OdkForm
from .admin import xformsubmit_return_message, xformsubmit_error
# odkdata modules
from odkdata import create_model, load_submit_data
from odkdata.utils import rm_digit, convert2camelcase


LOG = logging.getLogger(__name__)
LOG_DEBUG = logging.getLogger("mydebug")


def home(request):
    title = _("Welcome on django-odk")
    summary = _("Collection of geolocalized data.")

    return render(request, 'home.html', {'title': title, 'summary': summary})


class OdkGenView(object):

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.object:
            context['title'] = self.object.get_title()
        return context


class XFormTemplateView(generic.TemplateView):
    template_name = "xform_doc.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['title'] = _("XForm Documentation")
        return context


class XFormListView(generic.ListView):
    model = XForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['title'] = self.model._meta.verbose_name_plural
        return context    


class XFormDetailView(OdkGenView, generic.DetailView):
    model = XForm

    def get(self, request, *args, **kwargs):
        obj = self.get_object()

        if 'xls2xml' in request.path:
            obj.process_xform()
            obj.save()
            obj.set_xml_fields()
            obj.save()
            messages.success(request, _("XLSX successfully converted to XML XForm format."))
            return redirect(obj.get_absolute_url())
        if 'createmodel' in request.path:
            if create_model.process(obj):
                model_name = create_model.rm_digit(obj.form_id).capitalize()
                create_msg = _("created with success!")
                messages.success(request, f"Model odkdata.models.{model_name} {create_msg}")
                messages.info(request, _("You can now run 'manage.py makemigrations odkdata' then 'manage.py migrate'."))
                obj.model_created_on = now()
                obj.save()
                return redirect(obj.get_absolute_url())
            else:
                messages.error(request, "Error while creating model, check your error logs.")

        return super().get(request, *args, **kwargs)


@login_required(login_url='admin:login')
def xform_upload(request, pk=None):
    """Xform upload"""
    template = 'odk/xform_upload.html'
    uploaded_file_url = None
    obj = None
        
    if pk is not None:
        obj = XForm.objects.get(pk=pk)
        uploaded_file_url = obj.xml_file.url

    if request.method == 'POST' and request.FILES.get('xls_file'):
        file_pointer = request.FILES.get('xls_file')
        file_path = os.path.join("XForm", file_pointer.name)
        
        filename = overwrite_storage.save(file_path, file_pointer)

        uploaded_file_url = overwrite_storage.url(filename)

        obj = XForm.objects.create(xls_file=file_path, created_by=request.user, form_id=f"{filename} (tmp)")
        obj.save()
        # messages.success(request, _("xlsx file loaded"), fail_silently=True)
        return redirect(obj.get_absolute_url())

    context = {
        'form': OdkForm(data=request.POST, files=request.FILES, instance=obj),
        'title': _("XForm create & load"),
        'pk': pk,
        'uploaded_file_url': uploaded_file_url
    }

    return render(request, template, context)


class XFormDelView(LoginRequiredMixin, generic.DeleteView):
    model = XForm
    template_name = "odk/xform_detail.html"
    confirm_message = _("Delete this form?")
    success_url = reverse_lazy("odk:xform_list")

    def form_valid(self, form):
        """
        On delete rm xls and xml file
        """
        try:
            if self.object.xls_file:
                filePath = self.object.xls_file.path
                os.remove(filePath)
            if self.object.xml_file:
                filePath = self.object.xml_file.path
                os.remove(filePath)
        except Exception as e:
            msg = f"Error while deleting file {filePath}"
            LOG.error(msg)
            messages.warning(self.request, msg, fail_silently=True)
        self.object.delete()
        success_url = self.get_success_url()
        return HttpResponseRedirect(self.success_url)


class XFormSubmitDetailView(LoginRequiredMixin, OdkGenView, generic.DetailView):
    model = XFormSubmit

    def get(self, request, *args, **kwargs):
        obj = self.get_object()

        if 'loaddata/' in request.path:
            ok_list = []
            ko_list = []
            return_value, message = load_submit_data.load_record(obj)
            table_name = convert2camelcase(rm_digit(obj.form_id)).lower()
            if return_value < 0:
                xformsubmit_error(return_value, message, request, table_name)
            else:
                if return_value == 1:
                    ok_list.append(obj.id)
                    obj.inserted_on = now()
                    obj.save()
                else:
                    ko_list.append(obj.id)
                xformsubmit_return_message(ok_list, ko_list, message, request, table_name)

        return super().get(request, *args, **kwargs)   


# @login_required(login_url='admin:login')
def submittedfile_list(request):
    subpath = overwrite_storage.path("XFormSubmit")
    try:
        os.mkdir(subpath)
    except OSError as error:
        pass
    dirs, file_list = overwrite_storage.listdir("XFormSubmit")
    file_set = set(file_list)


    object_list = XFormSubmit.objects.all()
    db_set = set(XFormSubmit.objects.all().values_list('xml_file', flat=True))
    db_setok = {i.replace('XFormSubmit/', '') for i in db_set}

    missing_in_db = list(file_set - db_setok)
    context = {
        'object_list': object_list,
        'missing_in_db': missing_in_db,
        'title': XFormSubmit._meta.verbose_name
    }

    return render(request, 'odk/xformsubmit_list.html', context)


class XFormSubmitDelView(LoginRequiredMixin, generic.DeleteView):
    model = XFormSubmit
    template_name = "odk/xformsubmit_detail.html"
    confirm_message = _("Delete this submitted form?")
    success_url = reverse_lazy("odk:xformsubmit_list")

    def delete(self, request, *args, **kwargs):
        """
        On delete rm file
        """
        self.object = self.get_object()
        
        filePath = self.object.xml_file.path
        try:
            os.remove(filePath) # rm xml
        except:
            pass

        import glob
        filePictureDir = filePath.replace('.xml', '')
        pictures = glob.glob(f"{filePictureDir}/*", recursive=True)
        for pic in pictures:
            try:
                os.remove(pic)
            except OSError as xcpt:
                # msg = f"Error {xcpt} while deleting picture {pic}"
                # LOG.error(msg)
                return render(request, "500.html", {"error_msg": xcpt})
        os.rmdir(filePictureDir)

        success_url = self.get_success_url()
        self.object.delete()
        return HttpResponseRedirect(success_url)

  


"""
App storage to avoid multiple file creation on server like:
- sample_template.xml
- sample_template_AQXj3PD.xml
- sample_template_d2DxYEQ.xml
- sample_template_Pc45KRk.xml
"""
import os

from django.core.files.storage import FileSystemStorage
from config.settings import MEDIA_ROOT


class OverwriteMixin(object):
    """
    Update get_available_name to remove any previously stored file (if any) before returning the
    name.
    """
    def get_available_name(self, name, max_length=None):
        self.delete(name)
        return name


class OverwriteStorage(OverwriteMixin, FileSystemStorage):
    """A file-system based storage that let overwrite files with the same name."""


overwrite_storage = OverwriteStorage()
